
#include <string>
#include <fstream>
#include <iostream>

#include <Glob.hh>
#include <support.hh>
#include <os.hh>

#include <inttypes.h>

#include <usb-match.hh>

bool usb_module::operator== (const usb_module& dev) {
  
  const int USB_MATCH_VENDOR		= 0x0001;
  const int USB_MATCH_PRODUCT		= 0x0002;
  const int USB_MATCH_DEV_LO		= 0x0004;
  const int USB_MATCH_DEV_HI		= 0x0008;
  const int USB_MATCH_DEV_CLASS      	= 0x0010;
  const int USB_MATCH_DEV_SUBCLASS	= 0x0020;
  const int USB_MATCH_DEV_PROTOCOL	= 0x0040;
  const int USB_MATCH_INT_CLASS      	= 0x0080;
  const int USB_MATCH_INT_SUBCLASS	= 0x0100;
  const int USB_MATCH_INT_PROTOCOL	= 0x0200;
  
  if (match_flags & USB_MATCH_VENDOR && idVendor != dev.idVendor)
    return false;
  if (match_flags & USB_MATCH_PRODUCT && idProduct != dev.idProduct)
    return false;
  if (match_flags & USB_MATCH_DEV_LO && bcdDevice_lo != dev.bcdDevice_lo)
    return false;
  if (match_flags & USB_MATCH_DEV_HI && bcdDevice_hi != dev.bcdDevice_hi)
    return false;
  if (match_flags & USB_MATCH_DEV_CLASS && bDeviceClass != dev.bDeviceClass)
    return false;
  if (match_flags & USB_MATCH_DEV_SUBCLASS && bDeviceSubClass != dev.bDeviceSubClass)
    return false;
  if (match_flags & USB_MATCH_DEV_PROTOCOL && bDeviceProtocol != dev.bDeviceProtocol)
    return false;
  if (match_flags & USB_MATCH_INT_CLASS && bInterfaceClass != dev.bInterfaceClass)
    return false;
  if (match_flags & USB_MATCH_INT_SUBCLASS && bInterfaceSubClass != dev.bInterfaceSubClass)
    return false;
  if (match_flags & USB_MATCH_INT_PROTOCOL && bInterfaceProtocol != dev.bInterfaceProtocol)
    return false;
    
  return true;
};

void usb_module::read_from_sysfs (const std::string& file) {
  idVendor = read_value_from_file<uint32_t> (file + "/idVendor");
  idProduct = read_value_from_file<uint32_t> (file + "/idProduct");

  int bcdDevice = read_value_from_file<uint32_t> (file + "/bcdDevice");
  bcdDevice_lo = bcdDevice & 0xff;
  bcdDevice_hi = bcdDevice >> 8;

  bDeviceClass = read_value_from_file<uint32_t> (file + "/bDeviceClass");
  bDeviceSubClass = read_value_from_file<uint32_t> (file + "/bDeviceSubClass");
  bDeviceProtocol = read_value_from_file<uint32_t> (file + "/bDeviceProtocol");

  bInterfaceClass = read_value_from_file<uint32_t> (file + "/bInterfaceClass");
  bInterfaceSubClass = read_value_from_file<uint32_t> (file + "/bInterfaceSubClass");
  bInterfaceProtocol = read_value_from_file<uint32_t> (file + "/bInterfaceProtocol");
}

void usb_module::read_from_proc (const std::string& file) {
  std::ifstream f (file.c_str());
  
  unsigned char cs [64];
  f.read ((char*)cs, sizeof (cs));
  
  // in machine byte order ... 
  idVendor = *((uint16_t*)&cs[8]);
  idProduct = *((uint16_t*)&cs[10]);

  bcdDevice_lo = cs[14];
  bcdDevice_hi = cs[13];

  bDeviceClass = cs[4];
  bDeviceSubClass = cs[5];
  bDeviceProtocol = cs[6];

  bInterfaceClass = cs[32];
  bInterfaceSubClass = cs[33];
  bInterfaceProtocol = cs[34];
}

void usb_module::read_from_env ()
{
  // ACTION=%s [add or remove]
  // DEVPATH=%s [in 2.5 kernels, /sys/$DEVPATH]
  
  // we would also need a copy when we would use strtok in pure C ...
  std::string product = getenv ("PRODUCT"); // PRODUCT=%x/%x/%x
  std::string type = getenv ("TYPE"); // TYPE=%d/%d/%d
  std::string interface = getenv ("INTERFACE"); // INTERFACE=%d/%d/%d
  
  std::cout << product << " " << interface << " " << type << std::endl;

  /* when we have a stringstream and real << >> operator implementations
     in the embeddedSTL we can do this in nicer C++ style ... */
  
  std::string::size_type idx;
  
  idVendor = strtol(product.c_str(), NULL, 16);
  idx = product.find('/')+1;
  idProduct = strtol(product.c_str()+idx, NULL, 16);
  idx = product.find('/', idx)+1;
  int bcdDevice = strtol(product.c_str()+idx, NULL, 16);
  
  // this (inverted) works on the bigendian iBook - TODO check on x86
  bcdDevice_lo = bcdDevice >> 8;
  bcdDevice_hi = bcdDevice & 0xff;
  
  bDeviceClass = strtol(type.c_str(), NULL, 10);
  idx = type.find('/')+1;
  bDeviceSubClass = strtol(type.c_str()+idx, NULL, 10);
  idx = type.find('/', idx)+1;
  bDeviceProtocol = strtol(type.c_str()+idx, NULL, 10);
  
  bInterfaceClass = strtol(interface.c_str(), NULL, 10);
  idx = interface.find('/')+1;
  bInterfaceSubClass = strtol(interface.c_str()+idx, NULL, 10);
  idx = interface.find('/', idx)+1;
  bInterfaceProtocol = strtol(interface.c_str()+idx, NULL, 10);
  
  // DEVFS=/proc/bus/usb [gone in 2.5]
  // DEVICE=/proc/bus/usb/%03d/%03d
  std::cout << *this << std::endl;
}

std::ostream& operator<< (std::ostream& os, const usb_module& m) {
  os.setf (std::ios::hex, std::ios::basefield);
  
  os << "Vendor: " << m.idVendor << ", Product: " << m.idProduct
     << ", Dev_hi: " << m.bcdDevice_hi << ", Dev_lo: " << m.bcdDevice_lo
     << ", DeviceClass: " << m.bDeviceClass
     << ", DeviceSubClass: " << m.bDeviceSubClass
     << ", DeviceProtocol: " << m.bDeviceProtocol
     << ", InterfaceClass: " << m.bInterfaceClass
     << ", InterfaceSubClass: " << m.bInterfaceSubClass
     << ", InterfaceProtocol: " << m.bInterfaceProtocol;

  if (!m.module.empty())
    os << ": Module: " << m.module;

  return os;
}

// ...

std::vector<usb_module> usb_subsystem::modules;

void usb_subsystem::hotplug ()
{
  parse_modules();
  usb_module m;
  m.read_from_env ();
  match_modules (m, &usb_subsystem::hotplug);
}

void usb_subsystem::hotplug (const usb_module& m)
{
  if (m.module.size()) {
    modprobe (m.module);
  }
}

void usb_subsystem::hotplug_all ()
{
  parse_modules();
  for_each (&usb_subsystem::hotplug);
}

void usb_subsystem::list ()
{
  std::cout << "USB Devices and matched modules:" << std::endl;
  parse_modules();
  
  for_each (&usb_subsystem::list);
  std::cout << std::endl;
}

void usb_subsystem::list (const usb_module& m)
{
  std::cout << m << std::endl;
}

void usb_subsystem::match_modules (usb_module& m,
				   void(usb_subsystem::*op)(const usb_module&))
{
  std::vector<usb_module>::iterator module = modules.begin();
  while (module != modules.end()) {
    module = std::find (module, modules.end(), m);
    if (module != modules.end()) {
      m.module = module->module;
      (this->*op)(m);
      ++module;
    }
  }
}


void usb_subsystem::for_each (void(usb_subsystem::*op)(const usb_module&))
{
  if (0) {
  Utility::Glob usb_devices ("/sys/bus/usb/devices/*:*");
  for (Utility::Glob::iterator it = usb_devices.begin();
       it != usb_devices.end(); ++it)
    {
      usb_module m;
      m.read_from_sysfs (*it);
      match_modules (m, op);
    }
  if (usb_devices)
    return;
  }
  Utility::Glob usb_devices2 ("/proc/bus/usb/???""/???");
  for (Utility::Glob::iterator it = usb_devices2.begin();
       it != usb_devices2.end(); ++it)
    {
      usb_module m;
      m.read_from_proc (*it);
      match_modules (m, op);
    }
}

void usb_subsystem::parse_modules (const std::string &fname)
{
  std::ifstream f (fname.c_str());
  if (!f.is_open()) return;

  f.setf (std::ios::hex, std::ios::basefield);

  while (!f.eof()) {
    
    // quick hack to skip the leading comment line ...
    if (f.peek() == '#') {
      std::string temp;
      std::getline (f, temp);
      continue;
    }
    
    usb_module m;
    
    f >> m.module >> m.match_flags >> m.idVendor >> m.idProduct
      >> m.bcdDevice_lo >> m.bcdDevice_hi >> m.bDeviceClass >> m.bDeviceSubClass
      >> m.bDeviceProtocol >> m.bInterfaceClass >> m.bInterfaceSubClass
      >> m.bInterfaceProtocol >> m.driver_info;
    
    if (f.good())
      modules.push_back (m);
    
    if (f.fail() && !f.eof())
      f.clear(f.rdstate() & ~std::ios::failbit);
  }
}

void usb_subsystem::parse_modules ()
{
  if (modules.size())
    return;
  
  parse_modules("/lib/modules/" + kernel_release() + "/modules.usbmap");
  Utility::Glob maps("/etc/hotplug/usb.*map");
  for (Utility::Glob::iterator it = maps.begin(); it != maps.end(); ++it)
    parse_modules(*it);
}
